package org.eclipse.jdt.internal.compiler.classfmt;

import org.eclipse.jdt.internal.compiler.lookup.SignatureWrapper;
import org.eclipse.jdt.internal.compiler.env.IBinaryElementValuePair;
import org.eclipse.jdt.internal.compiler.env.IBinaryAnnotation;
import java.util.Iterator;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.env.ITypeAnnotationWalker;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import java.util.HashMap;
import java.io.LineNumberReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

public class ExternalAnnotationProvider
{
    public static final String ANNOTATION_FILE_EXTENSION = "eea";
    public static final String CLASS_PREFIX = "class ";
    public static final String SUPER_PREFIX = "super ";
    public static final char NULLABLE = '0';
    public static final char NONNULL = '1';
    public static final char NO_ANNOTATION = '@';
    public static final String ANNOTATION_FILE_SUFFIX = ".eea";
    private static final String TYPE_PARAMETER_PREFIX = " <";
    private String typeName;
    String typeParametersAnnotationSource;
    Map<String, String> supertypeAnnotationSources;
    private Map<String, String> methodAnnotationSources;
    private Map<String, String> fieldAnnotationSources;
    ExternalAnnotationProvider.SingleMarkerAnnotation NULLABLE_ANNOTATION;
    ExternalAnnotationProvider.SingleMarkerAnnotation NONNULL_ANNOTATION;
    
    public ExternalAnnotationProvider(InputStream input, String typeName) throws IOException {
        this.typeName = typeName;
        this.initialize(input);
    }
    
    private void initialize(InputStream input) throws IOException {
        Throwable t = null;
        try {
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));
            try {
                assertClassHeader(reader.readLine(), this.typeName);
                String line;
                if ((line = reader.readLine()) == null) {
                    return;
                }
                if (line.startsWith(" <")) {
                    if ((line = reader.readLine()) == null) {
                        return;
                    }
                    if (line.startsWith(" <")) {
                        this.typeParametersAnnotationSource = line.substring(" <".length());
                        if ((line = reader.readLine()) == null) {
                            return;
                        }
                    }
                }
                String pendingLine;
                do {
                    pendingLine = null;
                    line = line.trim();
                    Label_0536: {
                        if (!line.isEmpty()) {
                            String rawSig = null;
                            String annotSig = null;
                            String selector = line;
                            boolean isSuper = selector.startsWith("super ");
                            if (isSuper) {
                                selector = selector.substring("super ".length());
                            }
                            int errLine = -1;
                            try {
                                line = reader.readLine();
                                if (line != null && !line.isEmpty() && line.charAt(0) == ' ') {
                                    rawSig = line.substring(1);
                                }
                                else {
                                    errLine = reader.getLineNumber();
                                }
                                line = reader.readLine();
                                if (line != null) {
                                    if (line.isEmpty()) {
                                        break Label_0536;
                                    }
                                    if (line.charAt(0) != ' ') {
                                        pendingLine = line;
                                        break Label_0536;
                                    }
                                    annotSig = line.substring(1);
                                }
                                else {
                                    break Label_0536;
                                }
                            }
                            catch (Exception ex) {}
                            if (rawSig == null || annotSig == null) {
                                if (errLine == -1) {
                                    errLine = reader.getLineNumber();
                                }
                                throw new IOException("Illegal format for annotation file at line " + errLine);
                            }
                            else {
                                annotSig = trimTail(annotSig);
                                if (isSuper) {
                                    if (this.supertypeAnnotationSources == null) {
                                        this.supertypeAnnotationSources = new HashMap();
                                    }
                                    this.supertypeAnnotationSources.put(String.valueOf('L') + selector + rawSig + 59, annotSig);
                                }
                                else if (rawSig.contains("(")) {
                                    if (this.methodAnnotationSources == null) {
                                        this.methodAnnotationSources = new HashMap();
                                    }
                                    this.methodAnnotationSources.put(selector + rawSig, annotSig);
                                }
                                else {
                                    if (this.fieldAnnotationSources == null) {
                                        this.fieldAnnotationSources = new HashMap();
                                    }
                                    this.fieldAnnotationSources.put(selector + 58 + rawSig, annotSig);
                                }
                            }
                        }
                    }
                    line = pendingLine;
                } while (pendingLine != null || (line = reader.readLine()) != null);
            }
            finally {
                if (reader != null) {
                    reader.close();
                }
            }
        }
        finally {
            if (t == null) {
                Throwable exception;
                t = exception;
            }
            else {
                Throwable exception;
                if (t != exception) {
                    t.addSuppressed(exception);
                }
            }
        }
    }
    
    public static void assertClassHeader(String line, String typeName) throws IOException {
        if (line != null && line.startsWith("class ")) {
            line = line.substring("class ".length());
            if (!trimTail(line).equals(typeName)) {
                throw new IOException("mismatching class name in annotation file, expected " + typeName + ", but header said " + line);
            }
            return;
        }
        else {
            throw new IOException("missing class header in annotation file");
        }
    }
    
    public static String extractSignature(String line) {
        if (line == null || line.isEmpty() || line.charAt(0) != ' ') {
            return null;
        }
        return trimTail(line.substring(1));
    }
    
    protected static String trimTail(String line) {
        int tail = line.indexOf(32);
        if (tail == -1) {
            tail = line.indexOf(9);
        }
        if (tail != -1) {
            return line.substring(0, tail);
        }
        return line;
    }
    
    public ITypeAnnotationWalker forTypeHeader(LookupEnvironment environment) {
        if (this.typeParametersAnnotationSource != null || this.supertypeAnnotationSources != null) {
            return new ExternalAnnotationProvider.DispatchingAnnotationWalker(environment);
        }
        return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
    }
    
    public ITypeAnnotationWalker forMethod(char[] selector, char[] signature, LookupEnvironment environment) {
        Map<String, String> sources = this.methodAnnotationSources;
        if (sources != null) {
            String source = (String)sources.get(String.valueOf(CharOperation.concat(selector, signature)));
            if (source != null) {
                return new ExternalAnnotationProvider.MethodAnnotationWalker(source.toCharArray(), 0, environment);
            }
        }
        return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
    }
    
    public ITypeAnnotationWalker forField(char[] selector, char[] signature, LookupEnvironment environment) {
        if (this.fieldAnnotationSources != null) {
            String source = (String)this.fieldAnnotationSources.get(String.valueOf(CharOperation.concat(selector, signature, ':')));
            if (source != null) {
                return new ExternalAnnotationProvider.FieldAnnotationWalker(source.toCharArray(), 0, environment);
            }
        }
        return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("External Annotations for ").append(this.typeName).append('\n');
        sb.append("Methods:\n");
        if (this.methodAnnotationSources != null) {
            for (Map.Entry<String, String> e : this.methodAnnotationSources.entrySet()) {
                sb.append('\t').append((String)e.getKey()).append('\n');
            }
        }
        return sb.toString();
    }
    
    void initAnnotations(final LookupEnvironment environment) {
        if (this.NULLABLE_ANNOTATION == null) {
            this.NULLABLE_ANNOTATION = new ExternalAnnotationProvider.SingleMarkerAnnotation(this) {
                public char[] getTypeName() {
                    return this.getBinaryTypeName(environment.getNullableAnnotationName());
                }
            };
        }
        if (this.NONNULL_ANNOTATION == null) {
            this.NONNULL_ANNOTATION = new ExternalAnnotationProvider.SingleMarkerAnnotation(this) {
                public char[] getTypeName() {
                    return this.getBinaryTypeName(environment.getNonNullAnnotationName());
                }
            };
        }
    }
    
    public class TypeParametersAnnotationWalker extends ExternalAnnotationProvider.BasicAnnotationWalker
    {
        int[] rankStarts;
        int currentRank;
        
        TypeParametersAnnotationWalker(char[] source, int pos, int rank, int[] rankStarts, LookupEnvironment environment) {
            this$0.super(source, pos, environment);
            this.currentRank = rank;
            if (rankStarts != null) {
                this.rankStarts = rankStarts;
            }
            else {
                int length = source.length;
                rankStarts = new int[length];
                int curRank = 0;
                int depth = 0;
                boolean pendingVariable = true;
            Label_0306:
                for (int i = pos; i < length; ++i) {
                    switch (this.source[i]) {
                        case '<': {
                            ++depth;
                            break;
                        }
                        case '>': {
                            if (--depth < 0) {
                                break Label_0306;
                            }
                            break;
                        }
                        case ';': {
                            if (depth == 0 && i + 1 < length && this.source[i + 1] != ':') {
                                pendingVariable = true;
                                break;
                            }
                            break;
                        }
                        case ':': {
                            if (depth == 0) {
                                pendingVariable = true;
                            }
                            ++i;
                            while (i < length && this.source[i] == '[') {
                                ++i;
                            }
                            if (i < length && this.source[i] == 'L') {
                                for (int currentdepth = depth; i < length && (currentdepth != depth || this.source[i] != ';'); ++i) {
                                    if (this.source[i] == '<') {
                                        ++currentdepth;
                                    }
                                    if (this.source[i] == '>') {
                                        --currentdepth;
                                    }
                                }
                            }
                            --i;
                            break;
                        }
                        default: {
                            if (pendingVariable) {
                                pendingVariable = false;
                                rankStarts[curRank++] = i;
                                break;
                            }
                            break;
                        }
                    }
                }
                System.arraycopy(rankStarts, 0, this.rankStarts = new int[curRank], 0, curRank);
            }
        }
        
        @Override
        public ITypeAnnotationWalker toTypeParameter(boolean isClassTypeParameter, int rank) {
            if (rank == this.currentRank) {
                return this;
            }
            if (rank < this.rankStarts.length) {
                return this.this$0.new ExternalAnnotationProvider.TypeParametersAnnotationWalker(this.source, this.rankStarts[rank], rank, this.rankStarts, this.environment);
            }
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        @Override
        public ITypeAnnotationWalker toTypeParameterBounds(boolean isClassTypeParameter, int parameterRank) {
            return this.this$0.new ExternalAnnotationProvider.TypeParametersAnnotationWalker(this.source, this.rankStarts[parameterRank], parameterRank, this.rankStarts, this.environment);
        }
        
        @Override
        public ITypeAnnotationWalker toTypeBound(short boundIndex) {
            int p = this.pos;
            int i = this.currentTypeBound;
            while (true) {
                int colon = CharOperation.indexOf(':', this.source, p);
                if (colon != -1) {
                    p = colon + 1;
                }
                if (++i > boundIndex) {
                    break;
                }
                p = this.wrapperWithStart(p).computeEnd() + 1;
            }
            this.pos = p;
            this.currentTypeBound = boundIndex;
            return this;
        }
        
        public ITypeAnnotationWalker toField() {
            throw new UnsupportedOperationException("Cannot navigate to fields");
        }
        
        public ITypeAnnotationWalker toMethodReturn() {
            throw new UnsupportedOperationException("Cannot navigate to method return");
        }
        
        public ITypeAnnotationWalker toMethodParameter(short index) {
            throw new UnsupportedOperationException("Cannot navigate to method parameter");
        }
        
        public ITypeAnnotationWalker toThrows(int index) {
            throw new UnsupportedOperationException("Cannot navigate to throws");
        }
        
        @Override
        public IBinaryAnnotation[] getAnnotationsAtCursor(int currentTypeId) {
            if (this.pos != -1 && this.pos < this.source.length - 1) {
                switch (this.source[this.pos]) {
                    case '0': {
                        return new IBinaryAnnotation[] { this.this$0.NULLABLE_ANNOTATION };
                    }
                    case '1': {
                        return new IBinaryAnnotation[] { this.this$0.NONNULL_ANNOTATION };
                    }
                    default: {
                        break;
                    }
                }
            }
            return super.getAnnotationsAtCursor(currentTypeId);
        }
    }
    
    abstract class SingleMarkerAnnotation implements IBinaryAnnotation
    {
        public IBinaryElementValuePair[] getElementValuePairs() {
            return ElementValuePairInfo.NoMembers;
        }
        
        protected char[] getBinaryTypeName(char[][] name) {
            return CharOperation.concat('L', CharOperation.concatWith(name, '/'), ';');
        }
    }
    
    class DispatchingAnnotationWalker implements ITypeAnnotationWalker
    {
        private LookupEnvironment environment;
        private ExternalAnnotationProvider.TypeParametersAnnotationWalker typeParametersWalker;
        
        public DispatchingAnnotationWalker(LookupEnvironment environment) {
            this.environment = environment;
        }
        
        public ITypeAnnotationWalker toTypeParameter(boolean isClassTypeParameter, int rank) {
            String source = ExternalAnnotationProvider.this.typeParametersAnnotationSource;
            if (source != null) {
                if (this.typeParametersWalker == null) {
                    this.typeParametersWalker = new ExternalAnnotationProvider.TypeParametersAnnotationWalker(source.toCharArray(), 0, 0, null, this.environment);
                }
                return this.typeParametersWalker.toTypeParameter(isClassTypeParameter, rank);
            }
            else {
                return this;
            }
        }
        
        public ITypeAnnotationWalker toTypeParameterBounds(boolean isClassTypeParameter, int parameterRank) {
            if (this.typeParametersWalker != null) {
                return this.typeParametersWalker.toTypeParameterBounds(isClassTypeParameter, parameterRank);
            }
            return this;
        }
        
        public ITypeAnnotationWalker toSupertype(short index, char[] superTypeSignature) {
            Map<String, String> sources = ExternalAnnotationProvider.this.supertypeAnnotationSources;
            if (sources != null) {
                String source = (String)sources.get(String.valueOf(superTypeSignature));
                if (source != null) {
                    return new ExternalAnnotationProvider.SuperTypesAnnotationWalker(source.toCharArray(), this.environment);
                }
            }
            return this;
        }
        
        public ITypeAnnotationWalker toField() {
            return this;
        }
        
        public ITypeAnnotationWalker toThrows(int rank) {
            return this;
        }
        
        public ITypeAnnotationWalker toTypeArgument(int rank) {
            return this;
        }
        
        public ITypeAnnotationWalker toMethodParameter(short index) {
            return this;
        }
        
        public ITypeAnnotationWalker toTypeBound(short boundIndex) {
            return this;
        }
        
        public ITypeAnnotationWalker toMethodReturn() {
            return this;
        }
        
        public ITypeAnnotationWalker toReceiver() {
            return this;
        }
        
        public ITypeAnnotationWalker toWildcardBound() {
            return this;
        }
        
        public ITypeAnnotationWalker toNextArrayDimension() {
            return this;
        }
        
        public ITypeAnnotationWalker toNextNestedType() {
            return this;
        }
        
        public IBinaryAnnotation[] getAnnotationsAtCursor(int currentTypeId) {
            return ExternalAnnotationProvider.DispatchingAnnotationWalker.NO_ANNOTATIONS;
        }
    }
    
    abstract class BasicAnnotationWalker implements ITypeAnnotationWalker
    {
        char[] source;
        SignatureWrapper wrapper;
        int pos;
        int prevTypeArgStart;
        int currentTypeBound;
        LookupEnvironment environment;
        
        BasicAnnotationWalker(char[] source, int pos, LookupEnvironment environment) {
            this.source = source;
            this.pos = pos;
            ExternalAnnotationProvider.this.initAnnotations(this.environment = environment);
        }
        
        SignatureWrapper wrapperWithStart(int start) {
            if (this.wrapper == null) {
                this.wrapper = new SignatureWrapper(this.source);
            }
            this.wrapper.start = start;
            return this.wrapper;
        }
        
        public ITypeAnnotationWalker toReceiver() {
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toTypeParameter(boolean isClassTypeParameter, int rank) {
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toTypeParameterBounds(boolean isClassTypeParameter, int parameterRank) {
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toTypeBound(short boundIndex) {
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toSupertype(short index, char[] superTypeSignature) {
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toTypeArgument(int rank) {
            if (rank == 0) {
                int start = CharOperation.indexOf('<', this.source, this.pos) + 1;
                this.prevTypeArgStart = start;
                return new ExternalAnnotationProvider.MethodAnnotationWalker(this.source, start, this.environment);
            }
            int next = this.prevTypeArgStart;
            Label_0103: {
                switch (this.source[next]) {
                    case '*': {
                        break Label_0103;
                    }
                    case '+':
                    case '-': {
                        ++next;
                        break;
                    }
                }
                next = this.wrapperWithStart(next).computeEnd();
            }
            ++next;
            this.prevTypeArgStart = next;
            return new ExternalAnnotationProvider.MethodAnnotationWalker(this.source, next, this.environment);
        }
        
        public ITypeAnnotationWalker toWildcardBound() {
            switch (this.source[this.pos]) {
                case '+':
                case '-': {
                    return new ExternalAnnotationProvider.MethodAnnotationWalker(this.source, this.pos + 1, this.environment);
                }
                default: {
                    return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
                }
            }
        }
        
        public ITypeAnnotationWalker toNextArrayDimension() {
            if (this.source[this.pos] == '[') {
                int newPos = this.pos + 1;
                switch (this.source[newPos]) {
                    case '0':
                    case '1': {
                        ++newPos;
                        break;
                    }
                }
                return new ExternalAnnotationProvider.MethodAnnotationWalker(this.source, newPos, this.environment);
            }
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toNextNestedType() {
            return this;
        }
        
        public IBinaryAnnotation[] getAnnotationsAtCursor(int currentTypeId) {
            if (this.pos != -1 && this.pos < this.source.length - 2) {
                switch (this.source[this.pos]) {
                    case '*':
                    case '+':
                    case '-':
                    case 'L':
                    case 'T':
                    case '[': {
                        switch (this.source[this.pos + 1]) {
                            case '0': {
                                return new IBinaryAnnotation[] { ExternalAnnotationProvider.this.NULLABLE_ANNOTATION };
                            }
                            case '1': {
                                return new IBinaryAnnotation[] { ExternalAnnotationProvider.this.NONNULL_ANNOTATION };
                            }
                            default: {
                                break;
                            }
                        }
                    }
                }
            }
            return ExternalAnnotationProvider.BasicAnnotationWalker.NO_ANNOTATIONS;
        }
    }
    
    class SuperTypesAnnotationWalker extends ExternalAnnotationProvider.BasicAnnotationWalker
    {
        SuperTypesAnnotationWalker(char[] source, LookupEnvironment environment) {
            super(source, 0, environment);
        }
        
        public ITypeAnnotationWalker toField() {
            throw new UnsupportedOperationException("Supertype has no field annotations");
        }
        
        public ITypeAnnotationWalker toMethodReturn() {
            throw new UnsupportedOperationException("Supertype has no method return");
        }
        
        public ITypeAnnotationWalker toMethodParameter(short index) {
            throw new UnsupportedOperationException("Supertype has no method parameter");
        }
        
        public ITypeAnnotationWalker toThrows(int index) {
            throw new UnsupportedOperationException("Supertype has no throws");
        }
    }
    
    class MethodAnnotationWalker extends ExternalAnnotationProvider.BasicAnnotationWalker implements ExternalAnnotationProvider.IMethodAnnotationWalker
    {
        int prevParamStart;
        ExternalAnnotationProvider.TypeParametersAnnotationWalker typeParametersWalker;
        
        MethodAnnotationWalker(char[] source, int pos, LookupEnvironment environment) {
            super(source, pos, environment);
        }
        
        int typeEnd(int start) {
            while (this.source[start] == '[') {
                ++start;
                char an = this.source[start];
                if (an == '0' || an == '1') {
                    ++start;
                }
                else {
                    continue;
                }
            }
            int end = this.wrapperWithStart(start).computeEnd();
            return end;
        }
        
        @Override
        public ITypeAnnotationWalker toTypeParameter(boolean isClassTypeParameter, int rank) {
            if (this.source[0] == '<') {
                if (this.typeParametersWalker == null) {
                    return this.typeParametersWalker = new ExternalAnnotationProvider.TypeParametersAnnotationWalker(this.source, this.pos + 1, rank, null, this.environment);
                }
                return this.typeParametersWalker.toTypeParameter(isClassTypeParameter, rank);
            }
            else {
                return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
            }
        }
        
        @Override
        public ITypeAnnotationWalker toTypeParameterBounds(boolean isClassTypeParameter, int parameterRank) {
            if (this.typeParametersWalker != null) {
                return this.typeParametersWalker.toTypeParameterBounds(isClassTypeParameter, parameterRank);
            }
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toMethodReturn() {
            int close = CharOperation.indexOf(')', this.source);
            if (close != -1) {
                this.pos = close + 1;
                return this;
            }
            return ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER;
        }
        
        public ITypeAnnotationWalker toMethodParameter(short index) {
            if (index == 0) {
                int start = CharOperation.indexOf('(', this.source) + 1;
                this.prevParamStart = start;
                this.pos = start;
                return this;
            }
            int end = this.typeEnd(this.prevParamStart);
            ++end;
            this.prevParamStart = end;
            this.pos = end;
            return this;
        }
        
        public ITypeAnnotationWalker toThrows(int index) {
            return this;
        }
        
        public ITypeAnnotationWalker toField() {
            throw new UnsupportedOperationException("Methods have no fields");
        }
        
        @Override
        public int getParameterCount() {
            int count = 0;
            for (int start = CharOperation.indexOf('(', this.source) + 1; start < this.source.length && this.source[start] != ')'; start = this.typeEnd(start) + 1, ++count) {}
            return count;
        }
    }
    
    class FieldAnnotationWalker extends ExternalAnnotationProvider.BasicAnnotationWalker
    {
        public FieldAnnotationWalker(char[] source, int pos, LookupEnvironment environment) {
            super(source, pos, environment);
        }
        
        public ITypeAnnotationWalker toField() {
            return this;
        }
        
        public ITypeAnnotationWalker toMethodReturn() {
            throw new UnsupportedOperationException("Field has no method return");
        }
        
        public ITypeAnnotationWalker toMethodParameter(short index) {
            throw new UnsupportedOperationException("Field has no method parameter");
        }
        
        public ITypeAnnotationWalker toThrows(int index) {
            throw new UnsupportedOperationException("Field has no throws");
        }
    }
    
    public interface IMethodAnnotationWalker extends ITypeAnnotationWalker
    {
        public abstract int getParameterCount();
    }
}
